from distutils.core import setup

setup(
    name = 'HACNet',
    packages = ['HACNet'],
    version = '1.0.0',
    description = 'Python package',
    author = 'Gregory W. Kyro',
    author_email = 'gregory.kyro@yale.edu',
    url = 'https://github.com/gregory-kyro/HAC-Net/tree/main/HACNet',
    download_url = 'https://github.com/gregory-kyro/HAC-Net/archive/refs/tags/v1.3.tar.gz',
    keywords = [],
    classifiers = []
)
